"""Unit test helpers for Sentry instrumentation.

From the Sentry Python SDK's own unit tests:
https://github.com/getsentry/sentry-python/blob/c6a89d64db965fe0ece6de10df38ab936af8f5e4/tests/conftest.py
"""

import contextlib
from collections.abc import Callable, Generator
from dataclasses import dataclass
from typing import Any, override

import pytest
import sentry_sdk
from sentry_sdk.envelope import Envelope
from sentry_sdk.transport import Transport

__all__ = [
    "Attachment",
    "Captured",
    "TestTransport",
    "capture_events_fixture",
    "sentry_init_fixture",
]


@dataclass
class Attachment:
    """Contents and metadata of a Sentry attachment."""

    filename: str
    bytes: bytes
    content_type: str


@dataclass
class Captured:
    """A container for interesting items sent to Sentry."""

    errors: list[dict[str, Any]]
    transactions: list[dict[str, Any]]
    attachments: list[Attachment]


class TestTransport(Transport):
    """A transport that doesn't actually transport anything."""

    def __init__(self) -> None:
        Transport.__init__(self)

    @override
    def capture_envelope(self, envelope: Envelope) -> None:
        """No-op capture_envelope for tests."""


@contextlib.contextmanager
def sentry_init_fixture() -> Generator[Callable[..., None]]:
    """Return an init function that injects a no-op transport."""

    def inner(*a: Any, **kw: Any) -> None:
        kw.setdefault("transport", TestTransport())
        client = sentry_sdk.Client(*a, **kw)
        sentry_sdk.get_global_scope().set_client(client)

    old_client = sentry_sdk.get_global_scope().client
    try:
        sentry_sdk.get_current_scope().set_client(None)
        yield inner
    finally:
        sentry_sdk.get_global_scope().set_client(old_client)


def capture_events_fixture(
    monkeypatch: pytest.MonkeyPatch,
) -> Callable[[], Captured]:
    """Return a function that returns a container with items sent to Sentry."""

    def inner() -> Captured:
        test_client = sentry_sdk.get_client()
        if test_client.transport is None:
            raise RuntimeError(
                "Error patching Sentry transport: client.transport is None"
            )
        old_capture_envelope = test_client.transport.capture_envelope

        captured = Captured(errors=[], transactions=[], attachments=[])

        def append(envelope: Envelope) -> None:
            for item in envelope:
                match item.headers.get("type"):
                    case "event":
                        if item.payload.json is None:
                            raise ValueError(
                                "Sentry event unexpectedly missing json"
                                " payload"
                            )
                        captured.errors.append(item.payload.json)
                    case "transaction":
                        if item.payload.json is None:
                            raise ValueError(
                                "Sentry transaction unexpectedly missing json"
                                " payload"
                            )
                        captured.transactions.append(item.payload.json)
                    case "attachment":
                        captured.attachments.append(
                            Attachment(
                                filename=item.headers["filename"],
                                bytes=item.payload.get_bytes(),
                                content_type=item.headers["content_type"],
                            )
                        )
                    case _:
                        pass
            return old_capture_envelope(envelope)

        monkeypatch.setattr(test_client.transport, "capture_envelope", append)

        return captured

    return inner



"""Test metrics tooling."""

from __future__ import annotations

import asyncio
import math
from datetime import timedelta
from enum import Enum
from typing import cast
from uuid import UUID

import pytest
from aiokafka import AIOKafkaConsumer
from aiokafka.admin.client import AIOKafkaAdminClient, NewTopic
from dataclasses_avroschema.pydantic import AvroBaseModel
from faststream.kafka import KafkaBroker
from pydantic import Field
from schema_registry.client.client import AsyncSchemaRegistryClient
from schema_registry.serializers.message_serializer import (
    AsyncAvroMessageSerializer,
)

from safir.dependencies.metrics import EventDependency, EventMaker
from safir.kafka import (
    KafkaConnectionSettings,
    PydanticSchemaManager,
    SchemaManagerSettings,
)
from safir.metrics import (
    DisabledMetricsConfiguration,
    DuplicateEventError,
    EventManager,
    EventManagerUnintializedError,
    EventMetadata,
    EventPayload,
    EventPublisher,
    EventsConfiguration,
    KafkaEventManager,
    KafkaEventPublisher,
    KafkaMetricsConfiguration,
    KafkaTopicError,
)


class MyEvent(EventPayload):
    """An event payload."""

    foo: str
    duration: timedelta
    duration_union: timedelta | None


class Events(EventMaker):
    """A class to hold event publishers."""

    async def initialize(self, manager: EventManager) -> None:
        self.my_event = await manager.create_publisher("myevent", MyEvent)


async def subscribe_and_wait(consumer: AIOKafkaConsumer, pattern: str) -> None:
    """Avoid race condition when subscribing to topics.

    There is some race condition where an exception is raised if we start
    trying to consume messages too soon after we have subscribed to a topic
    or pattern, because partitions haven't been assigned to the consumer yet.
    """
    consumer.subscribe(pattern=pattern)
    await asyncio.sleep(0.5)


async def assert_from_kafka(
    consumer: AIOKafkaConsumer,
    schema_registry: AsyncSchemaRegistryClient,
    event: EventPublisher[MyEvent],
    foo: str,
    duration: timedelta,
) -> None:
    """Grab a message from kafka, deserialize it, and assert stuff about it."""
    message = await consumer.getone()
    assert isinstance(message.value, bytes)
    serializer = AsyncAvroMessageSerializer(schema_registry)
    deserialized_dict = await serializer.decode_message(message.value)
    assert deserialized_dict is not None
    event_class = cast("type[AvroBaseModel]", event._event_class)
    deserialized = event_class(**deserialized_dict)

    assert isinstance(deserialized, EventMetadata)
    assert isinstance(deserialized.id, UUID)
    assert deserialized.application == "testapp"

    # dataclasses-avroschema serializes python datetime's into avro
    # timestamp-millis's
    deserialized_ms = math.trunc(deserialized.timestamp_ns / 1e6)
    assert deserialized_ms == deserialized.timestamp.timestamp() * 1e3

    # Mypy can't deal with the fact that deserialized has both EventMetadata
    # AND MyEvent as bases, so let's pretend we don't know either :(
    # https://github.com/python/mypy/issues/12314
    assert getattr(deserialized, "foo", None) == foo
    assert getattr(deserialized, "duration", None) == duration


async def integration_test(
    manager: EventManager,
    schema_manager_settings: SchemaManagerSettings,
    kafka_consumer: AIOKafkaConsumer,
    kafka_admin_client: AIOKafkaAdminClient,
    *,
    create_topic: bool = True,
) -> None:
    """Test the interaction of the EventManager and the EventsDependency."""
    schema_registry = AsyncSchemaRegistryClient(
        **schema_manager_settings.to_registry_params()
    )

    if create_topic:
        topic = NewTopic(
            name="what.ever.testapp",
            num_partitions=1,
            replication_factor=1,
        )
        await kafka_admin_client.create_topics([topic])

    # Initialize the event manager
    await manager.initialize()

    # Create an events dependency and initialize it with the transport config
    events_dependency = EventDependency(Events())
    await events_dependency.initialize(manager=manager)

    # Publish events
    event = events_dependency.events.my_event
    await event.publish(
        MyEvent(
            foo="bar1",
            duration=timedelta(seconds=2, milliseconds=123),
            duration_union=None,
        )
    )
    published = await event.publish(
        MyEvent(
            foo="bar2",
            duration=timedelta(seconds=2, milliseconds=456),
            duration_union=None,
        )
    )
    schema = published.avro_schema_to_python()
    assert schema["name"] == "myevent"
    assert schema["namespace"] == "what.ever.testapp"

    # Set up a kafka consumer
    expected_topic = "what.ever.testapp"
    await subscribe_and_wait(consumer=kafka_consumer, pattern=expected_topic)
    await kafka_consumer.seek_to_beginning()

    # Assert stuff
    assert isinstance(event, KafkaEventPublisher)
    assert event._publisher.topic == expected_topic
    await assert_from_kafka(
        kafka_consumer,
        schema_registry,
        event,
        "bar1",
        timedelta(seconds=2.123),
    )
    await assert_from_kafka(
        kafka_consumer,
        schema_registry,
        event,
        "bar2",
        timedelta(seconds=2.456),
    )

    await manager.aclose()


@pytest.mark.asyncio
async def test_managed_storage(
    kafka_connection_settings: KafkaConnectionSettings,
    schema_manager_settings: SchemaManagerSettings,
    kafka_consumer: AIOKafkaConsumer,
    kafka_admin_client: AIOKafkaAdminClient,
) -> None:
    """Publish events to actual storage and read them back and verify them."""
    config = KafkaMetricsConfiguration(
        application="testapp",
        events=EventsConfiguration(topic_prefix="what.ever"),
        kafka=kafka_connection_settings,
        schema_manager=schema_manager_settings,
    )

    # Construct an event manager and intialize our events dependency
    manager = config.make_manager()
    await integration_test(
        manager, schema_manager_settings, kafka_consumer, kafka_admin_client
    )

    # Make sure storage is cleaned up
    assert manager._admin_client._closed
    assert not await manager._broker.ping(timeout=1)


@pytest.mark.asyncio
async def test_unmanaged_storage(
    schema_manager_settings: SchemaManagerSettings,
    kafka_consumer: AIOKafkaConsumer,
    kafka_broker: KafkaBroker,
    kafka_admin_client: AIOKafkaAdminClient,
    schema_manager: PydanticSchemaManager,
) -> None:
    """Publish events to actual storage and read them back and verify them."""
    manager = KafkaEventManager(
        application="testapp",
        topic_prefix="what.ever",
        kafka_broker=kafka_broker,
        kafka_admin_client=kafka_admin_client,
        schema_manager=schema_manager,
        manage_kafka=False,
    )
    await integration_test(
        manager, schema_manager_settings, kafka_consumer, kafka_admin_client
    )

    # Make sure storage is NOT cleaned up
    assert not manager._admin_client._closed
    assert await manager._broker.ping(timeout=1)


@pytest.mark.asyncio
async def test_topic_not_created(
    schema_manager_settings: SchemaManagerSettings,
    kafka_consumer: AIOKafkaConsumer,
    kafka_broker: KafkaBroker,
    kafka_admin_client: AIOKafkaAdminClient,
    schema_manager: PydanticSchemaManager,
) -> None:
    manager = KafkaEventManager(
        application="testapp",
        topic_prefix="what.ever",
        kafka_broker=kafka_broker,
        kafka_admin_client=kafka_admin_client,
        schema_manager=schema_manager,
        manage_kafka=False,
    )

    with pytest.raises(KafkaTopicError):
        await integration_test(
            manager,
            schema_manager_settings,
            kafka_consumer,
            kafka_admin_client,
            create_topic=False,
        )


@pytest.mark.asyncio
async def test_create_before_initialize(
    kafka_broker: KafkaBroker,
    kafka_admin_client: AIOKafkaAdminClient,
    schema_manager: PydanticSchemaManager,
) -> None:
    topic = NewTopic(
        name="what.ever.testapp",
        num_partitions=1,
        replication_factor=1,
    )
    await kafka_admin_client.create_topics([topic])

    manager = KafkaEventManager(
        application="testapp",
        topic_prefix="what.ever",
        kafka_broker=kafka_broker,
        kafka_admin_client=kafka_admin_client,
        schema_manager=schema_manager,
        manage_kafka=False,
    )

    with pytest.raises(EventManagerUnintializedError):
        await manager.create_publisher("myevent", MyEvent)


@pytest.mark.asyncio
async def test_duplicate_event(event_manager: EventManager) -> None:
    class MyOtherEvent(EventPayload):
        blah: int

    await event_manager.create_publisher("myevent", MyEvent)
    with pytest.raises(DuplicateEventError):
        await event_manager.create_publisher("myevent", MyOtherEvent)


@pytest.mark.asyncio
async def test_invalid_payload(event_manager: EventManager) -> None:
    class MyEnum(Enum):
        case1 = "case1"
        case2 = "case2"

    class MyInvalidEvent(EventPayload):
        good_simple_field: str = Field()
        good_union_field: str | None = Field()
        good_enum_field: MyEnum = Field()
        bad_list_field: list[str] = Field()
        bad_union_field: dict[str, str] | None = Field()
        bad_dict_field: dict[str, str] = Field()

    with pytest.raises(ValueError, match="Unsupported Avro Schema") as excinfo:
        await event_manager.create_publisher("myinvalidevent", MyInvalidEvent)
    err = str(excinfo.value)

    assert "bad_list_field" in err
    assert "bad_union_field" in err
    assert "bad_dict_field" in err

    assert "good_simple_field" not in err
    assert "good_union_field" not in err
    assert "good_enum_field" not in err


@pytest.mark.asyncio
async def test_disable() -> None:
    config = DisabledMetricsConfiguration(
        enabled=False,
        application="testapp",
        events=EventsConfiguration(topic_prefix="what.ever"),
    )
    manager = config.make_manager()

    await manager.initialize()
    event = await manager.create_publisher("myevent", MyEvent)

    await event.publish(
        MyEvent(foo="bar1", duration=timedelta(seconds=1), duration_union=None)
    )
    await event.publish(
        MyEvent(foo="bar2", duration=timedelta(seconds=1), duration_union=None)
    )
    await manager.aclose()



"""Helpers for tests that require a Confluent schema registry."""

from typing import Any, Self

import httpx
from httpx import ReadError, RemoteProtocolError
from testcontainers.core.container import DockerContainer
from testcontainers.core.network import Network
from testcontainers.core.waiting_utils import wait_container_is_ready

from ..constants import SCHEMA_REGISTRY_DOCKER_IMAGE

__all__ = ["SchemaRegistryContainer"]


class SchemaRegistryContainer(DockerContainer):
    """A ``Testcontainers <https://github.com/testcontainers/testcontainers-python>``
    Confluent schema registry container.

    Parameters
    ----------
    network
        A Docker network to put this container in.
        The kafa container that will serve as storage must be on the same
        network.
    kafka_bootstrap_servers
        A string with a comma-separated list of kafka boostrap servers for the
        kafka instance that will serve as storage
    image
        The docker image to use.
    """

    def __init__(
        self,
        network: Network,
        kafka_bootstrap_servers: str = "kafka:9092",
        image: str = SCHEMA_REGISTRY_DOCKER_IMAGE,
        **kwargs: dict[str, Any],
    ) -> None:
        super().__init__(image, **kwargs)
        self.port = 8081
        self.with_exposed_ports(self.port)
        self.with_env(
            "SCHEMA_REGISTRY_KAFKASTORE_BOOTSTRAP_SERVERS",
            kafka_bootstrap_servers,
        )
        self.with_env("SCHEMA_REGISTRY_LISTENERS", "http://0.0.0.0:8081")
        self.with_env(
            "SCHEMA_REGISTRY_HOST_NAME", self.get_container_host_ip()
        )

    def start(self, *args: list[Any], **kwargs: dict[str, Any]) -> Self:
        super().start(*args, **kwargs)
        self.health()
        return self

    def get_url(self) -> str:
        host = self.get_container_host_ip()
        port = self.get_exposed_port(self.port)
        return f"http://{host}:{port}"

    def reset(self) -> None:
        url = f"{self.get_url()}/subjects"
        subjects = httpx.get(url).json()
        for subject in subjects:
            httpx.delete(f"{url}/{subject}")
            httpx.delete(f"{url}/{subject}?permanent=true")

    @wait_container_is_ready(ReadError, RemoteProtocolError)
    def health(self) -> None:
        """We're health when we can be queried from the host network."""
        url = f"{self.get_url()}/subjects"
        httpx.get(url, timeout=5).raise_for_status()



"""Test fixtures."""

from __future__ import annotations

from collections.abc import AsyncGenerator, Generator, Iterator
from datetime import timedelta
from pathlib import Path

import pytest
import pytest_asyncio
import respx
from aiokafka import AIOKafkaConsumer
from aiokafka.admin.client import AIOKafkaAdminClient, NewTopic
from faststream.kafka import KafkaBroker
from pydantic import AnyUrl
from redis.asyncio import Redis
from testcontainers.core.container import Network
from testcontainers.postgres import PostgresContainer
from testcontainers.redis import RedisContainer

from safir.kafka import (
    KafkaConnectionSettings,
    PydanticSchemaManager,
    SchemaManagerSettings,
    SecurityProtocol,
)
from safir.metrics import (
    EventManager,
    EventsConfiguration,
    KafkaMetricsConfiguration,
)
from safir.sentry import (
    before_send_handler,
    fingerprint_env_handler,
    sentry_exception_handler,
)
from safir.testing.gcs import MockStorageClient, patch_google_storage
from safir.testing.kubernetes import MockKubernetesApi, patch_kubernetes
from safir.testing.sentry import (
    Captured,
    capture_events_fixture,
    sentry_init_fixture,
)
from safir.testing.slack import MockSlackWebhook, mock_slack_webhook

from .support.kafka.container import FullKafkaContainer
from .support.schema_registry import SchemaRegistryContainer


@pytest.fixture(scope="session")
def kafka_docker_network() -> Iterator[Network]:
    """Provide a network object to link session-scoped testcontainers."""
    with Network() as network:
        yield network


@pytest.fixture(scope="session")
def kafka_cert_path(
    tmp_path_factory: pytest.TempPathFactory,
) -> Path:
    return tmp_path_factory.mktemp("kafka-certs")


@pytest.fixture(scope="session")
def global_kafka_container(
    kafka_docker_network: Network,
    kafka_cert_path: Path,
) -> Iterator[FullKafkaContainer]:
    """Provide a session-scoped kafka container.

    You proably want one of the dependent test-scoped fixtures that clears
    kafka data, like:
    * ``kafka_container``
    * ``kafka_broker``
    * ``kafka_consumer``
    * ``kafka_admin_client``
    """
    container = FullKafkaContainer(limit_broker_to_first_host=True)
    container.with_network(kafka_docker_network)
    container.with_network_aliases("kafka")
    with container as kafka:
        (kafka_cert_path / "ca.crt").write_text(
            container.get_secret_file_contents("ca.crt")
        )
        (kafka_cert_path / "client.crt").write_text(
            container.get_secret_file_contents("client.crt")
        )
        (kafka_cert_path / "client.key").write_text(
            container.get_secret_file_contents("client.key")
        )
        yield kafka


@pytest.fixture
def kafka_container(
    global_kafka_container: FullKafkaContainer,
) -> Iterator[FullKafkaContainer]:
    """Yield the global kafka container, but rid it of data post-test."""
    yield global_kafka_container
    global_kafka_container.reset()


@pytest.fixture
def kafka_connection_settings(
    kafka_container: FullKafkaContainer,
) -> KafkaConnectionSettings:
    """Provide a url to a session-scoped kafka container.

    All data is cleared from the kafka instance at the end of the test.
    """
    return KafkaConnectionSettings(
        bootstrap_servers=kafka_container.get_bootstrap_server(),
        security_protocol=SecurityProtocol.PLAINTEXT,
    )


@pytest_asyncio.fixture
async def kafka_consumer(
    kafka_connection_settings: KafkaConnectionSettings,
) -> AsyncGenerator[AIOKafkaConsumer]:
    """Provide an AOIKafkaConsumer pointed at a session-scoped kafka container.

    All data is cleared from the kafka instance at the end of the test.
    """
    consumer = AIOKafkaConsumer(
        **kafka_connection_settings.to_aiokafka_params(),
        client_id="pytest-consumer",
    )
    await consumer.start()
    yield consumer
    await consumer.stop()


@pytest_asyncio.fixture
async def kafka_broker(
    kafka_connection_settings: KafkaConnectionSettings,
) -> AsyncGenerator[KafkaBroker]:
    """Provide a fast stream KafkaBroker pointed at a session-scoped kafka
    container.

    All data is cleared from the kafka instance at the end of the test.
    """
    broker = KafkaBroker(
        **kafka_connection_settings.to_faststream_params(),
        client_id="pytest-broker",
    )
    await broker.start()
    yield broker
    await broker.close()


@pytest_asyncio.fixture
async def kafka_admin_client(
    kafka_connection_settings: KafkaConnectionSettings,
) -> AsyncGenerator[AIOKafkaAdminClient]:
    """Provide an AOIKafkaAdmin client pointed at a session-scoped kafka
    container.

    All data is cleared from the kafka instance at the end of the test.
    """
    client = AIOKafkaAdminClient(
        **kafka_connection_settings.to_aiokafka_params(),
        client_id="pytest-admin",
    )
    await client.start()
    yield client
    await client.close()


@pytest.fixture(scope="session")
def global_schema_registry_container(
    global_kafka_container: FullKafkaContainer,
    kafka_docker_network: Network,
) -> Iterator[SchemaRegistryContainer]:
    """Provide a session-scoped schema registry container that can talk to the
    containers in other docker-based kafka fixtures.

    You probably want one of the dependent test-scoped fixtures that clears
    registry data, like ``schema_registry_connection_settings`` or
    ``schema_manager``.
    """
    container = SchemaRegistryContainer(network=kafka_docker_network)
    container.with_network(kafka_docker_network)
    container.with_network_aliases("schemaregistry")
    with container as schema_registry:
        yield schema_registry


@pytest.fixture
def schema_registry_container(
    global_schema_registry_container: SchemaRegistryContainer,
) -> Iterator[SchemaRegistryContainer]:
    """Yield the global schema registry container and rid of data post-test."""
    yield global_schema_registry_container
    global_schema_registry_container.reset()


@pytest.fixture
def schema_manager_settings(
    schema_registry_container: SchemaRegistryContainer,
) -> SchemaManagerSettings:
    """Provide a URL to a session-scoped schema registry.

    All data is cleared from it at the end of the test.
    """
    return SchemaManagerSettings(
        registry_url=AnyUrl(schema_registry_container.get_url())
    )


@pytest.fixture
def schema_manager(
    schema_manager_settings: SchemaManagerSettings,
) -> PydanticSchemaManager:
    """Provide a PydanticSchemaManager pointed at a session-scoped schema
    registry container.

    All data is cleared from the registry at the end of the test.
    """
    return schema_manager_settings.make_manager()


@pytest_asyncio.fixture
async def event_manager(
    kafka_connection_settings: KafkaConnectionSettings,
    schema_manager_settings: SchemaManagerSettings,
) -> AsyncGenerator[EventManager]:
    """Provide an event manager and create a matching Kafka topic."""
    config = KafkaMetricsConfiguration(
        application="testapp",
        enabled=True,
        events=EventsConfiguration(topic_prefix="what.ever"),
        kafka=kafka_connection_settings,
        schema_manager=schema_manager_settings,
    )

    manager = config.make_manager()
    await manager.initialize()
    topic = NewTopic(
        name="what.ever.testapp",
        num_partitions=1,
        replication_factor=1,
    )
    await manager._admin_client.create_topics([topic])
    yield manager
    await manager.aclose()


@pytest.fixture(scope="session")
def database_password() -> str:
    return "INSECURE@%PASSWORD/"


@pytest.fixture(scope="session")
def database_url(database_password: str) -> Iterator[str]:
    """Start a PostgreSQL database and return a URL for it."""
    with PostgresContainer(
        driver="asyncpg", username="safir", password=database_password
    ) as postgres:
        yield postgres.get_connection_url()


@pytest.fixture
def mock_gcs() -> Iterator[MockStorageClient]:
    yield from patch_google_storage(
        expected_expiration=timedelta(hours=1), bucket_name="some-bucket"
    )


@pytest.fixture
def mock_kubernetes() -> Iterator[MockKubernetesApi]:
    yield from patch_kubernetes()


@pytest.fixture
def mock_slack(respx_mock: respx.Router) -> MockSlackWebhook:
    return mock_slack_webhook("https://example.com/slack", respx_mock)


@pytest.fixture(scope="session")
def redis() -> Iterator[RedisContainer]:
    """Start a Redis container."""
    with RedisContainer() as redis:
        yield redis


@pytest_asyncio.fixture
async def redis_client(redis: RedisContainer) -> AsyncGenerator[Redis]:
    """Create a Redis client for testing.

    This must be done separately for each test since it's tied to the per-test
    event loop, and therefore must be separated from the session-shared Redis
    server container.
    """
    host = redis.get_container_host_ip()
    port = redis.get_exposed_port(6379)
    client = Redis(host=host, port=port, db=0)
    yield client
    await client.aclose()


@pytest.fixture
def sentry_fingerprint_items(
    monkeypatch: pytest.MonkeyPatch,
) -> Generator[Captured]:
    """Mock sentry transport and add env to event fingerprints."""
    with sentry_init_fixture() as init:
        init(
            environment="some_env",
            traces_sample_rate=1.0,
            before_send=fingerprint_env_handler,
        )
        events = capture_events_fixture(monkeypatch)
        yield events()


@pytest.fixture
def sentry_exception_items(
    monkeypatch: pytest.MonkeyPatch,
) -> Generator[Captured]:
    """Mock sentry transport and add SentryException info."""
    with sentry_init_fixture() as init:
        init(
            environment="some_env",
            traces_sample_rate=1.0,
            before_send=sentry_exception_handler,
        )
        events = capture_events_fixture(monkeypatch)
        yield events()


@pytest.fixture
def sentry_combo_items(
    monkeypatch: pytest.MonkeyPatch,
) -> Generator[Captured]:
    """Mock sentry transport and add all recommended before_send processing."""
    with sentry_init_fixture() as init:
        init(
            environment="some_env",
            traces_sample_rate=1.0,
            before_send=before_send_handler,
        )
        events = capture_events_fixture(monkeypatch)
        yield events()



"""Constants for metrics functionality."""

BROKER_PREFIX = "safir-metrics-faststream-broker"
"""Prefix for the Kafka client id for the metrics FastStream broker."""

ADMIN_CLIENT_PREFIX = "safir-metrics-admin-client"
"""Prefix for the client id for the metrics Kafka admin client."""



"""Metrics exceptions."""

__all__ = [
    "DuplicateEventError",
    "EventManagerUnintializedError",
    "KafkaTopicError",
]


class EventManagerUnintializedError(Exception):
    """An attempt to create a publisher after manager has been initialized."""


class DuplicateEventError(Exception):
    """Two publishers were registered with the same name."""

    def __init__(self, name: str) -> None:
        super().__init__(
            f"{name}: you have already created an event with this "
            " name. Events must have unique names within this "
            " application."
        )


class KafkaTopicError(Exception):
    """A topic does not exist in Kafka, or we don't have access to it."""

    def __init__(self, topic: str) -> None:
        super().__init__(
            f"Topic: {topic} does not exist in Kafka, or we don't have the"
            " permissions to access it. See https://safir.lsst.io/metrics"
            " for how to provision Kafka topics, access, and other "
            " infrastructure for publishing metrics events."
        )



"""Kafka Testcontainer with listeners for different security protocols.

This is the testcontainers built-in module:
https://github.com/testcontainers/testcontainers-python/blob/main/modules/kafka/testcontainers/kafka/__init__.py

With these differences:
* Provision SSL, SASL_SSL, and SASL_PLAINTEXT listeners in addition to the
  existing PLAINTEXT listener.
* Optionally mount a host directory for the generated kafka ssl secrets so
  they can be used in test clients
* Expose the SASL username and password with ``.get_sasl_username()`` and
  ``.get_sasl_password()``
* Provide ``.reset()`` to delete all topics from the running kafka instance
* Optionally control ``limit_broker_to_first_host`` with an init var
  instead of an env var
* Always uses Kraft, no more Zookeeper
* Type annotation fixes
* Lint fixes
"""

import tarfile
import time
from dataclasses import dataclass, field
from io import BytesIO
from os import environ
from pathlib import Path
from textwrap import dedent
from typing import Any, Self

from testcontainers.core.container import DockerContainer
from testcontainers.core.utils import raise_for_deprecated_parameter
from testcontainers.core.version import ComparableVersion
from testcontainers.core.waiting_utils import wait_for_logs

__all__ = [
    "FullKafkaContainer",
    "kafka_config",
]
LIMIT_BROKER_ENV_VAR = "TC_KAFKA_LIMIT_BROKER_TO_FIRST_HOST"


@dataclass
class _KafkaConfig:
    limit_broker_to_first_host: bool = field(
        default_factory=lambda: environ.get(LIMIT_BROKER_ENV_VAR) == "true"
    )
    """
    This option is useful for a setup with a network,
    see testcontainers/testcontainers-python#637 for more details
    """


kafka_config = _KafkaConfig()


class FullKafkaContainer(DockerContainer):
    """
    Kafka container.

    Example:

        .. doctest::

            >>> from testcontainers.kafka import KafkaContainer

            >>> with KafkaContainer() as kafka:
            ...     connection = kafka.get_bootstrap_server()
            ...

            # Using KRaft protocol
            >>> with KafkaContainer().with_kraft() as kafka:
            ...     connection = kafka.get_bootstrap_server()
            ...
    """  # fmt: skip

    TC_START_SCRIPT = "/tc-start.sh"
    MIN_KRAFT_TAG = "7.0.0"

    def __init__(
        self,
        image: str = "confluentinc/cp-kafka:7.6.0",
        port: int = 9093,
        ssl_port: int = 29093,
        sasl_plaintext_port: int = 29094,
        sasl_ssl_port: int = 29095,
        limit_broker_to_first_host: bool | None = None,
        **kwargs: Any,
    ) -> None:
        raise_for_deprecated_parameter(kwargs, "port_to_expose", "port")
        super().__init__(image, **kwargs)

        self.limit_broker_to_first_host = (
            limit_broker_to_first_host
            if limit_broker_to_first_host is not None
            else kafka_config.limit_broker_to_first_host
        )
        self.container_cert_path = Path("/etc/kafka/secrets")
        self.host_cert_script_path = Path(__file__).parent / "data"
        self.container_cert_script_path = Path(
            "/var/generate-kafka-certs.bash"
        )
        self.with_volume_mapping(
            host=str(self.host_cert_script_path),
            container="/var/testcontainers-init",
            mode="ro",
        )

        self.port = port
        self.sasl_plaintext_port = sasl_plaintext_port
        self.ssl_port = ssl_port
        self.sasl_ssl_port = sasl_ssl_port
        self.with_exposed_ports(
            self.port,
            self.ssl_port,
            self.sasl_plaintext_port,
            self.sasl_ssl_port,
        )

        self.sasl_username = "admin"
        self.sasl_password = "admin"

        # Kraft
        self.wait_for = r".*\[KafkaServer id=\d+\] started.*"
        self.boot_command = ""
        self.cluster_id = "MkU3OEVBNTcwNTJENDM2Qk"
        self.listeners = (
            f"PLAINTEXT://0.0.0.0:{self.port},BROKER://0.0.0.0:9092,"
            f"SSL://0.0.0.0:{self.ssl_port},"
            f"SASL_SSL://0.0.0.0:{self.sasl_ssl_port},"
            f"SASL_PLAINTEXT://0.0.0.0:{self.sasl_plaintext_port}"
        )

        self.security_protocol_map = (
            "BROKER:PLAINTEXT,PLAINTEXT:PLAINTEXT,SSL:SSL,SASL_SSL:SASL_SSL,"
            "SASL_PLAINTEXT:SASL_PLAINTEXT"
        )

        self.with_env("KAFKA_LISTENERS", self.listeners)
        self.with_env(
            "KAFKA_LISTENER_SECURITY_PROTOCOL_MAP", self.security_protocol_map
        )
        self.with_env("KAFKA_INTER_BROKER_LISTENER_NAME", "BROKER")

        self.with_env("KAFKA_BROKER_ID", "1")
        self.with_env("KAFKA_OFFSETS_TOPIC_REPLICATION_FACTOR", "1")
        self.with_env("KAFKA_OFFSETS_TOPIC_NUM_PARTITIONS", "1")
        self.with_env("KAFKA_LOG_FLUSH_INTERVAL_MESSAGES", "10000000")
        self.with_env("KAFKA_GROUP_INITIAL_REBALANCE_DELAY_MS", "0")

        # SSL
        self.with_env("KAFKA_SSL_KEYSTORE_FILENAME", "server.keystore.jks")
        self.with_env("KAFKA_SSL_KEYSTORE_CREDENTIALS", "credentials")
        self.with_env("KAFKA_SSL_KEY_CREDENTIALS", "credentials")
        self.with_env(
            "KAFKA_SSL_TRUSTSTORE_FILENAME",
            "server.truststore.jks",
        )
        self.with_env("KAFKA_SSL_TRUSTSTORE_CREDENTIALS", "credentials")

        # SASL
        self.with_env("KAFKA_SASL_ENABLED_MECHANISMS", "SCRAM-SHA-512")
        self.with_env(
            "KAFKA_OPTS",
            f"-Djava.security.auth.login.config={self.container_cert_path!s}/kafka_server_jaas.conf",
        )

        self._verify_min_kraft_version()
        self.kraft_enabled = True

    def _verify_min_kraft_version(self) -> None:
        actual_version = self.image.split(":")[-1]

        if ComparableVersion(actual_version) < self.MIN_KRAFT_TAG:
            raise ValueError(
                f"Provided Confluent Platform's version {actual_version} "
                f"is not supported in Kraft mode"
                f" (must be {self.MIN_KRAFT_TAG} or above)"
            )

    def with_cluster_id(self, cluster_id: str) -> Self:
        self.cluster_id = cluster_id
        return self

    def configure(self) -> None:
        self.wait_for = r".*Kafka Server started.*"

        self.with_env("CLUSTER_ID", self.cluster_id)
        self.with_env("KAFKA_NODE_ID", 1)
        self.with_env(
            "KAFKA_LISTENER_SECURITY_PROTOCOL_MAP",
            f"{self.security_protocol_map},CONTROLLER:PLAINTEXT",
        )
        self.with_env(
            "KAFKA_LISTENERS",
            f"{self.listeners},CONTROLLER://0.0.0.0:9094",
        )
        self.with_env("KAFKA_PROCESS_ROLES", "broker,controller")

        network_alias = self._get_network_alias()
        controller_quorum_voters = f"1@{network_alias}:9094"
        self.with_env(
            "KAFKA_CONTROLLER_QUORUM_VOTERS", controller_quorum_voters
        )
        self.with_env("KAFKA_CONTROLLER_LISTENER_NAMES", "CONTROLLER")

        self.boot_command = f"""
                    sed -i '/KAFKA_ZOOKEEPER_CONNECT/d' \
                      /etc/confluent/docker/configure
                    echo 'kafka-storage format --ignore-formatted \
                      -t {self.cluster_id} -c /etc/kafka/kafka.properties' \
                      >> /etc/confluent/docker/configure
                """

    def _get_network_alias(self) -> str | None:
        if self._network:
            return next(
                iter(
                    self._network_aliases
                    or [self._network.name or self._kwargs.get("network", [])]
                ),
                None,
            )

        return "localhost"

    def get_bootstrap_server(self) -> str:
        host = self.get_container_host_ip()
        port = self.get_exposed_port(self.port)
        return f"{host}:{port}"

    def get_ssl_bootstrap_server(self) -> str:
        host = self.get_container_host_ip()
        port = self.get_exposed_port(self.ssl_port)
        return f"{host}:{port}"

    def get_sasl_ssl_bootstrap_server(self) -> str:
        host = self.get_container_host_ip()
        port = self.get_exposed_port(self.sasl_ssl_port)
        return f"{host}:{port}"

    def get_sasl_plaintext_bootstrap_server(self) -> str:
        host = self.get_container_host_ip()
        port = self.get_exposed_port(self.sasl_plaintext_port)
        return f"{host}:{port}"

    def get_sasl_username(self) -> str:
        return self.sasl_username

    def get_sasl_password(self) -> str:
        return self.sasl_password

    def tc_start(self) -> None:
        host = self.get_container_host_ip()
        port = self.get_exposed_port(self.port)
        ssl_port = self.get_exposed_port(self.ssl_port)
        sasl_ssl_port = self.get_exposed_port(self.sasl_ssl_port)
        sasl_plaintext_port = self.get_exposed_port(self.sasl_plaintext_port)

        if self.limit_broker_to_first_host:
            listeners = (
                f"PLAINTEXT://{host}:{port},"
                f"BROKER://$(hostname -i | cut -d' ' -f1):9092,"
                f"SSL://{host}:{ssl_port},SASL_SSL://{host}:{sasl_ssl_port},"
                f"SASL_PLAINTEXT://{host}:{sasl_plaintext_port}"
            )
        else:
            listeners = (
                f"PLAINTEXT://{host}:{port},"
                f"BROKER://$(hostname -i):9092,SSL://{host}:{port},"
                f"SASL_SSL://{host}:{sasl_ssl_port},"
                f"SASL_PLAINTEXT://{host}:{sasl_plaintext_port}"
            )
        data = (
            dedent(
                f"""
                #!/bin/bash
                /var/testcontainers-init/generate-kafka-secrets.bash \
                  {self.container_cert_path!s} {host}
                {self.boot_command}
                export KAFKA_ADVERTISED_LISTENERS={listeners}
                . /etc/confluent/docker/bash-config
                /etc/confluent/docker/configure
                /etc/confluent/docker/launch
                """
            )
            .strip()
            .encode("utf-8")
        )
        self.create_file(data, self.TC_START_SCRIPT)

    def start(self, timeout: int = 30) -> Self:
        script = self.TC_START_SCRIPT
        command = (
            f'sh -c "while [ ! -f {script} ]; do sleep 0.1; done; sh {script}"'
        )
        self.configure()
        self.with_command(command)
        super().start()
        self.tc_start()
        wait_for_logs(self, self.wait_for, timeout=timeout)
        self.exec(
            f"kafka-configs --bootstrap-server localhost:9092 --alter "
            f"--add-config 'SCRAM-SHA-512=[iterations=8192,"
            f"password={self.sasl_password}]' --entity-type users "
            f"--entity-name {self.sasl_username}"
        )
        return self

    def get_secret_file_contents(self, name: str) -> str:
        """Get the contents of a file from /etc/kafka/secrets.

        Useful for getting TLS certs and keys for connecting with the SSL
        security protocol.

        We can't just mount a host volume over /etc/kafka/secrets because the
        OS user that owns that directory could be different than that
        ``appuser`` user that runs kafka in the container.
        """
        raw = BytesIO()
        path = Path("/etc/kafka/secrets/") / name
        bits, _ = self.get_wrapped_container().get_archive(str(path))
        for chunk in bits:
            raw.write(chunk)
        raw.seek(0)
        with tarfile.open(fileobj=raw) as tf:
            stream = tf.extractfile(name)
            assert stream
            data = stream.read()
            assert data
            return data.decode()

    def reset(self) -> None:
        self.exec(
            "/bin/kafka-topics --bootstrap-server localhost:9092 --delete"
            " --topic '.*'"
        )

    def create_file(self, content: bytes, path: str) -> None:
        with (
            BytesIO() as archive,
            tarfile.TarFile(fileobj=archive, mode="w") as tar,
        ):
            tarinfo = tarfile.TarInfo(name=path)
            tarinfo.size = len(content)
            tarinfo.mtime = int(time.time())
            tar.addfile(tarinfo, BytesIO(content))
            archive.seek(0)
            self.get_wrapped_container().put_archive("/", archive)
